/* This is a quick hack as I lost the mood to work on my current big
 * project, which seems to last for months/years... :-(
 *
 * Actually it's a port from my Fast-Fractal-Package done for the
 * Sinclair-QL between April 1989 and July 1992.
 *
 * Note:
 *   This program is *not* meant to be portable in that sense, as it relies
 *   upon the SVGA Library and uses i80486 assembly code to speed things up.
 *   (although the assembly code should run fine on a i80386 too)
 *
 * Now, compile and have fun.
 *
 * Any comments and/or suggestions are welcome.
 *
 * This little piece of software is Freeware.
 * Do what you want with it, except stating you've written it.  THX
 *
 * Thomas M. Ott
 * eMail: Labalutsch@aol.com
 */

#define	VERSION		"v1.2.2 @1997 Apr 28 by Th.M. Ott"
#define	VERSION_LAST	"v1.2.1 @1997 Mar 04 by Th.M. Ott"

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include STRING_HEADER
#include VARARG_HEADER

#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <time.h>

#include <vga.h>


struct mode_info {
    const int  x, y;
    int	       mode;
  };

struct keep {
    struct mode_info  *mp;
    unsigned char     *sbuf;
  };

struct cmd_line_par {
    int	    iter;
    double  rmin, rmax, imin, imax;
  };


void
  asm_draw_mandel P( (int, int, int,
		      int, int, int, int) ),
  asm_draw_julia P( (int, int, int,
		     int, int, int, int, int, int) );

static void
  fp_draw_mandel P( (int, int, int,
		     double, double, double, double) ),
  fp_draw_julia P( (int, int, int,
		    double, double, double, double, double, double) ),
  init_mode P( (struct mode_info *) ),
  driver P( (const int, struct cmd_line_par *) ),
  keep_image P( (struct keep *) ),
  last_image P( (struct keep *) ),
  print_time P( (int, clock_t) ),
  save_image P( (char *, int, double, double, double, double) ),
  wait4user P( (const char *) );

static int
  do_mandel P( (int, int,
		double, double, double, double) ),
  do_julia P( (int, int,
	       double, double, double, double) ),
  do_zoom P( (int (*) P( (int, int, double, double, double, double) ),
	      int, int, double, double, double, double) ),
  setup_julia P( (int, double, double, double, double) ),
  get_iter_count P( (int) ),
  get_X_value P( (int) ),
  get_Y_value P( (int) );

static unsigned char
  change_resolution P( (void) );

static __NORETURN void
  error P( (const int, const char *, ...) ) __NORETURN2;

static __NORETURN INLINE void
  give_help P( (void) ) __NORETURN2;


#ifndef	TRUE
  #define  FALSE	0
  #define  TRUE		!FALSE
#endif

#define	GAME_RESO	"game"
#define	LOW_RESO	"low"
#define	MID_RESO	"mid"
#define	HIGH_RESO	"high"
#define	FP_ONLY		"fp-only"

#define	OPT_HELP	"help"
#define	OPT_VERSION	"version"

#define	GAME_X_RESO	320
#define	GAME_Y_RESO	200

#define	LOW_X_RESO	640
#define	LOW_Y_RESO	480

#define	MID_X_RESO	800
#define	MID_Y_RESO	600

#define	HIGH_X_RESO	1024
#define	HIGH_Y_RESO	768

#define	MAX_X_RESO	HIGH_X_RESO
#define	MAX_Y_RESO	HIGH_Y_RESO


static struct mode_info
  mode_tab[]= {
      { GAME_X_RESO, GAME_Y_RESO, -1 },
      { LOW_X_RESO,  LOW_Y_RESO,  -1 },
      { MID_X_RESO,  MID_Y_RESO,  -1 },
      { HIGH_X_RESO, HIGH_Y_RESO, -1 }
    },
  *modep= &mode_tab[ 1];

#define	MAX_MODES	( sizeof( mode_tab)/ sizeof( mode_tab[ 0]))

static char
  *progname;

static double
  jreal_con= 0.0,
  jimag_con= 0.0;

static unsigned char
  fp_only= FALSE;


static __NORETURN INLINE void
give_help()
{
  printf( "Usage: %s [OPTIONS] "
	  "[iter real_min real_max imag_min imag_max "
	  "[real_const imag_const]]\n\n"
	  "  -g, --%s\tgame resolution 320x200\n"
	  "  -l, --%s\tlow resolution 640x480 (default)\n"
	  "  -m, --%s\tmid resolution 800x600\n"
	  "  -h, --%s\thigh resolution 1024x768\n"
	  "  -f, --%s\talways use floating-point arithmetic\n"
	  "      --%s\tprint this help and exit\n"
	  "      --%s\tprint version information and exit\n\n"
	  "  Typing `h' invokes the on-line help\n"
	  "  The man-page gives more detailed information\n\n",
	  progname,
	  GAME_RESO, LOW_RESO, MID_RESO, HIGH_RESO, FP_ONLY,
	  OPT_HELP, OPT_VERSION);
  __NORETURN exit( EXIT_SUCCESS);
}


int
main( argc, argv)
  int	argc;
  char	**argv;
{
  struct cmd_line_par  params, *parp;
  double	       *dp;
  char		       *p, buf[ 30], *endp;
  int		       mode, i;

  p= strrchr( *argv, '/');
  progname= p == NULL ? *argv : p+ 1;
  while ( --argc != 0)
    {
      p= *++argv;
      if ( *p++ != '-')
	break;
      if ( *p++ == '-')
	{
	  if ( *p == '\0')
	    {
	      ++argv;
	      --argc;
	      break;
	    }
	  if ( strcmp( GAME_RESO, p) == 0)
	    modep= &mode_tab[ 0];
	  else if ( strcmp( LOW_RESO, p) == 0)
	    modep= &mode_tab[ 1];
	  else if ( strcmp( MID_RESO, p) == 0)
	    modep= &mode_tab[ 2];
	  else if ( strcmp( HIGH_RESO, p) == 0)
	    modep= &mode_tab[ 3];
	  else if ( strcmp( FP_ONLY, p) == 0)
	    fp_only= TRUE;
	  else if ( strcmp( OPT_HELP, p) == 0)
	    __NORETURN give_help();
	  else if ( strcmp( OPT_VERSION, p) == 0)
	    {
	      printf( "%s %s\n", progname, VERSION);
	      __NORETURN exit( EXIT_SUCCESS);
	    }
	  else
	    __NORETURN error( TRUE, "%s `--%s'", "unknown option", p);
	  continue;
	}
      if ( *--p == '\0')
	break;
      do
	switch ( *p++)
	  {
	    default:
	      __NORETURN error( TRUE, "%s `-%c'", "unknown option", *--p);
	    case 'g':
	      modep= &mode_tab[ 0];
	      continue;
	    case 'l':
	      modep= &mode_tab[ 1];
	      continue;
	    case 'm':
	      modep= &mode_tab[ 2];
	      continue;
	    case 'h':
	      modep= &mode_tab[ 3];
	      continue;
	    case 'f':
	      fp_only= TRUE;
	      continue;
	  }
      while ( *p != '\0');
    }
  if ( argc == 0)
    parp= NULL;
  else if ( argc != 5 && argc != 7)
    __NORETURN error( TRUE, "got %d instead of 5 or 7 parameters", argc);
  else
    {
      parp= &params;
      parp->iter= strtol( *argv, &endp, 10);
      if ( endp == *argv || *endp != '\0' || parp->iter < 0)
	__NORETURN error( TRUE, "invalid %s parameter `%s'",
			  "iteration", *argv);
      for ( dp= &parp->rmin, p= "real minreal maximag minimag max";
	    *p != '\0';
	    p += 8, dp++)
	{
	  *dp= strtod( *++argv, &endp);
	  if ( endp == *argv || *endp != '\0')
	    __NORETURN error( TRUE, "invalid %.8s parameter `%s'", p, *argv);
	  if ( fabs( *dp) > 3.999999999)
	    __NORETURN error( FALSE, "parameter `%.8s' out of range "
				     "( %.15f > %1.9f)",
			      p, *dp, 3.999999999);
	}
      if ( ( p -= 32, parp->rmin >= parp->rmax)
	   || ( p += 16, parp->imin >= parp->imax))
	__NORETURN error( TRUE, "parameter `%.8s' must be lower than `%.8s'",
			  p, p+ 8);
      if ( ( argc -= 5) != 0)
	for ( dp= &jreal_con, p= "realimag";
	      *p != '\0';
	      p += 4, dp= &jimag_con)
	  {
	    *dp= strtod( *++argv, &endp);
	    if ( endp == *argv || *endp != '\0')
	      __NORETURN error( TRUE, "invalid %.4s constant parameter `%s'",
				p, *argv);
	    if ( fabs( *dp) > 3.999999999)
	      __NORETURN error( TRUE, "parameter `%.4s constant' out of "
				      "range ( %1.15f > %1.9f)",
				p, *dp, 3.999999999);
	  }
    }
  vga_init();
  for ( i= 0;  i < MAX_MODES;  i++)
    {
      sprintf( buf, "G%dx%dx256", mode_tab[ i].x, mode_tab[ i].y);
      mode= vga_getmodenumber( buf);
      if ( mode >= 0 && mode <= vga_lastmodenumber())
	mode_tab[ i].mode= mode;
    }
  if ( modep->mode < 0)
    __NORETURN error( FALSE, "mode `%s' %s", buf, "unavailable");
  init_mode( modep);
  driver( argc != 0, parp);
  vga_setmode( TEXT);
  return( EXIT_SUCCESS);
}


static __NORETURN void
error( const int need_help, const char *fmt, ...)
{
  va_list  par;

  fprintf( stderr, "%s: ", progname);
  va_start( par, fmt);
  vfprintf( stderr, fmt, par);
  va_end( par);
  fputc( '\n', stderr);
  if ( need_help)
    fprintf( stderr, "Try `%s --%s' for more information\n",
	     progname, OPT_HELP);
  __NORETURN exit( EXIT_FAILURE);
}


#define	FIX_SCALE	( 1 << 29)
#define	dbl2fix( x)	(int) ( (x)* FIX_SCALE)

  /* --- Mandelbrot section --- */

#define	MANDEL_INT_ZOOM	27500.0

#define	M_REAL_MIN	-2.0
#define	M_REAL_MAX	1.0

#define	M_IMAG_MIN	-1.25
#define	M_IMAG_MAX	1.25

#define	M_ITER		64

#define	M_REAL_SCR	( M_REAL_MAX- M_REAL_MIN)
#define	M_IMAG_SCR	( M_IMAG_MAX- M_IMAG_MIN)


static void
driver( julia_flag, parp)
  const int	       julia_flag;
  struct cmd_line_par  *parp;
{
  int  res;

  if ( parp == NULL)
    res= 'r';
  else
    {
      printf( "computing command line parameters...\n");
      res= julia_flag
	? do_julia ( 1, parp->iter,
		       parp->rmin, parp->rmax, parp->imin, parp->imax)
	: do_mandel( 1, parp->iter,
		     parp->rmin, parp->rmax, parp->imin, parp->imax);
    }
  while ( res != 'q')
    res= do_mandel( 1, M_ITER, M_REAL_MIN, M_REAL_MAX, M_IMAG_MIN, M_IMAG_MAX);
}


#define	GO_BACK_MSG	"go back to image"


static int
do_mandel( stage, iter, real_min, real_max, imag_min, imag_max)
  int	  stage, iter;
  double  real_min, real_max, imag_min, imag_max;
{
  static unsigned char
    use_int_mandel= TRUE;

  double       real_step, imag_step, zoom_real, zoom_imag;
  clock_t      st;
  int	       ch;
  struct keep  keep;

  zoom_real= M_REAL_SCR/ fabs( real_max- real_min);
  zoom_imag= M_IMAG_SCR/ fabs( imag_max- imag_min);
  keep.sbuf= NULL;
  for ( ; ; )
    {
      if ( keep.sbuf != NULL)
	last_image( &keep);
      else
	{
	  real_step= fabs( real_max- real_min)/ modep->x;
	  imag_step= fabs( imag_max- imag_min)/ modep->y;
	  vga_setcolor( 0);
	  vga_clear();
	  if ( fp_only
	       || zoom_real > MANDEL_INT_ZOOM || zoom_imag > MANDEL_INT_ZOOM)
	    {
	      if ( use_int_mandel)
		{
		  printf( "\a%s switching over to `fp' math", "Mandel:");
		  fflush( stdout);
		  use_int_mandel= FALSE;
		}
	      st= clock();
	      fp_draw_mandel( iter, modep->x- 1, modep->y- 1,
			      real_max, imag_min, real_step, imag_step);
	    }
	  else
	    {
	      if ( !use_int_mandel)
		{
		  printf( "\a%s falling back to `int' math", "Mandel:");
		  fflush( stdout);
		  use_int_mandel= TRUE;
		}
	      st= clock();
	      asm_draw_mandel( iter, modep->x- 1, modep->y- 1,
			       dbl2fix( real_max), dbl2fix( imag_min),
			       dbl2fix( real_step), dbl2fix( imag_step));
	    }
	  st= clock()- st;
	}
      printf( "\a\n%s", "Mandel:");
      printf( "\treal( %1.9f, %1.9f), imag( %1.9f, %1.9f),\n"
	      "\tzoom( %.2f, %.2f), %d iterations using `%s' math\n",
	      real_min, real_max, imag_min, imag_max,
	      zoom_real, zoom_imag, iter, use_int_mandel ? "int" : "fp");
      print_time( stage, st);
    repeat:
      switch ( ch= vga_getch())
	{
	  default:
	    fputc( '\a', stdout);
	    fflush( stdout);
	    goto  repeat;
	  case 'z':
	    ch= do_zoom( do_mandel, stage, iter,
			 real_min, real_max, imag_min, imag_max);
	    if ( ch == 'b')
	      continue;
	    if ( ch == 'l')
	      if ( keep.sbuf != NULL)
		continue;
	      else
		break;
	    if ( ch <= '9' && ch >= '1')
	      {
	  case '1':  case '2':  case '3':  case '4':  case '5':
	  case '6':  case '7':  case '8':  case '9':
		if ( --ch == '0')
		  ch= 'b';
	      }
	  case 'q':
	  case 'b':
	  case 'r':
	  case 'l':
	    break;
	  case 'j':
	    ch= setup_julia( iter, real_min, real_max, imag_min, imag_max);
	    if ( ch == 'b' || ( ch == 'l' && keep.sbuf != NULL))
	      continue;
	    break;
	  case 's':
	    save_image( "Mandel:", iter,
			real_min, real_max, imag_min, imag_max);
	    goto  repeat;
	  case 'k':
	    keep_image( &keep);
	    goto  repeat;
	  case 'h':
	    printf( "\nMandelbrot Set Help\n===================\n\n"
		    "  `z':\t\tzoom deeper into Mandelbrot set\n"
		    "  `j':\t\tselect Julia set\n"
		    "  `r':\t\trestart Mandelbrot set\n"
		    "  `b' or `1':\tgo back to previous Mandelbrot zoom stage\n"
		    "  `2' .. `9':\tgo back N Mandelbrot zoom stages\n");
	    printf( "  `k':\t\tkeep image save and mark it for later restoring\n"
		    "  `l':\t\tgo back to last marked image\n"
		    "  `s':\t\tsave current image to file\n"
		    "  `c':\t\tchange resolution\n"
		    "  `q':\t\texit program\n"
		    "  'h':\t\tthis help screen\n\n");
	  case '\t':
	    wait4user( GO_BACK_MSG);
	    goto  repeat;
	  case 'c':
	    if ( change_resolution())
	      continue;
	    goto  repeat;
	}
      break;
    }
  if ( keep.sbuf != NULL)
    free( keep.sbuf);
  return( ch);
}


static void
fp_draw_mandel( itc, upx, y, real_max, imag_val, real_step, imag_step)
  int	  itc, upx, y;
  double  real_max, imag_val, real_step, imag_step;
{
  int	  iter, x;
  double  real_val, real, imag, real_and_imag, square_sum;

  /* for a description of what's going on here, have a look at the
   * assembly file `mandel.S'
   */
  real_max += real_step;
  imag_val -= imag_step;
  do
    {
      imag_val += imag_step;
      real_val= real_max;
      x= upx;
      do
	{
	  real_val -= real_step;
	  real= real_val;
	  imag= imag_val;
	  for ( iter= 1- itc;  ;  )
	    {
	      real_and_imag= real+ imag;
	      if ( ( real *= real) >= 4.0 || ( imag *= imag) >= 4.0
		   || ( square_sum= real+ imag) >= 4.0)
		{
		  iter= ( ( iter+ itc) & 127)+ 1;
		  break;
		}
	      if ( ++iter == 0)
		break;
	      real += real_val- imag;
	      imag= real_and_imag* real_and_imag- square_sum+ imag_val;
	    }
	  vga_setcolor( iter);
	  vga_drawpixel( x, y);
	}
      while ( --x >= 0);
    }
  while ( --y >= 0);
}

  /* --- Julia section --- */

#define	JULIA_INT_ZOOM	435.0

#define	J_REAL_MIN	-1.85
#define	J_REAL_MAX	1.85

#define	J_IMAG_MIN	-1.35
#define	J_IMAG_MAX	1.35

#define	J_REAL_SCR	( J_REAL_MAX- J_REAL_MIN)
#define	J_IMAG_SCR	( J_IMAG_MAX- J_IMAG_MIN)


static int
setup_julia( iter, real_min, real_max, imag_min, imag_max)
  int	  iter;
  double  real_min, real_max, imag_min, imag_max;
{
  static int
    x= LOW_X_RESO/ 2,
    y= LOW_Y_RESO/ 2;

  int	  res;
  double  scale;

  x= get_X_value( x < modep->x ? x : modep->x/ 2);
  y= get_Y_value( y < modep->y ? y : modep->y/ 2);
  scale= fabs( real_max- real_min)/ modep->x;
  jreal_con= real_min+ scale* x;
  scale= fabs( imag_max- imag_min)/ modep->y;
  jimag_con= imag_max- scale* y;
  do
    {
      res= do_julia( 1, iter, J_REAL_MIN, J_REAL_MAX, J_IMAG_MIN, J_IMAG_MAX);
      if ( res == 'R')
	return( 'r');
      else if ( res == 'm')
	return( 'b');
    }
  while ( res != 'q' && res != 'l');
  return( res);
}


static int
do_julia( stage, iter, real_min, real_max, imag_min, imag_max)
  int	  stage, iter;
  double  real_min, real_max, imag_min, imag_max;
{
  static unsigned char
    use_int_julia= TRUE;

  double       real_step, imag_step, zoom_real, zoom_imag;
  clock_t      st;
  int	       ch;
  struct keep  keep;

  zoom_real= J_REAL_SCR/ fabs( real_max- real_min);
  zoom_imag= J_IMAG_SCR/ fabs( imag_max- imag_min);
  keep.sbuf= NULL;
  for ( ; ; )
    {
      if ( keep.sbuf != NULL)
	last_image( &keep);
      else
	{
	  real_step= fabs( real_max- real_min)/ modep->x;
	  imag_step= fabs( imag_max- imag_min)/ modep->y;
	  vga_setcolor( 0);
	  vga_clear();
	  if ( fp_only
	       || zoom_real > JULIA_INT_ZOOM || zoom_imag > JULIA_INT_ZOOM)
	    {
	      if ( use_int_julia)
		{
		  printf( "\a%s switching over to `fp' math", "Julia:");
		  fflush( stdout);
		  use_int_julia= FALSE;
		}
	      st= clock();
	      fp_draw_julia( iter, modep->x- 1, modep->y- 1,
			     real_max, imag_min, real_step, imag_step,
			     jreal_con, jimag_con);
	    }
	  else
	    {
	      if ( !use_int_julia)
		{
		  printf( "\a%s falling back to `int' math", "Julia:");
		  fflush( stdout);
		  use_int_julia= TRUE;
		}
	      st= clock();
	      asm_draw_julia( iter, modep->x- 1, modep->y- 1,
			      dbl2fix( real_max), dbl2fix( imag_min),
			      dbl2fix( real_step), dbl2fix( imag_step),
			      dbl2fix( jreal_con), dbl2fix( jimag_con));
	    }
	  st= clock()- st;
	}
      printf( "\a\n%s constant( %1.9f, %1.9f),\n",
	      "Julia:", jreal_con, jimag_con);
      printf( "\treal( %1.9f, %1.9f), imag( %1.9f, %1.9f),\n"
	      "\tzoom( %.2f, %.2f), %d iterations using `%s' math\n",
	      real_min, real_max, imag_min, imag_max,
	      zoom_real, zoom_imag, iter, use_int_julia ? "int" : "fp");
      print_time( stage, st);
    repeat:
      switch ( ch= vga_getch())
	{
	  default:
	    fputc( '\a', stdout);
	    fflush( stdout);
	    goto  repeat;
	  case 'z':
	    ch= do_zoom( do_julia, stage, iter,
			 real_min, real_max, imag_min, imag_max);
	    if ( ch == 'b')
	      continue;
	    if ( ch == 'l')
	      if ( keep.sbuf != NULL)
		continue;
	      else
		break;
	    if ( ch <= '9' && ch >= '1')
	      {
	  case '1':  case '2':  case '3':  case '4':  case '5':
	  case '6':  case '7':  case '8':  case '9':
		if ( --ch == '0')
		  ch= 'b';
	      }
	  case 'q':
	  case 'b':
	  case 'r':
	  case 'm':
	  case 'R':
	  case 'l':
	    break;
	  case 's':
	    save_image( "Julia:", iter,
			real_min, real_max, imag_min, imag_max);
	    goto  repeat;
	  case 'k':
	    keep_image( &keep);
	    goto  repeat;
	  case 'h':
	    printf( "\nJulia Set Help\n==============\n\n"
		    "  `z':\t\tzoom deeper into Julia set\n"
		    "  `r':\t\trestart Julia set\n"
		    "  `R':\t\trestart Mandelbrot set\n"
		    "  `m':\t\tback to invoking Mandelbrot set\n"
		    "  `b' or `1':\tgo back to previous Julia zoom stage\n"
		    "  `2' .. `9':\tgo back `N' Julia zoom stages\n");
	    printf( "  `k':\t\tkeep image save and mark it for later restoring\n"
		    "  'l':\t\tgo back to last marked image\n"
		    "  `s':\t\tsave current image to file\n"
		    "  `c':\t\tchange resolution\n"
		    "  `q':\t\texit program\n"
		    "  'h':\t\tthis help screen\n\n");
	  case '\t':
	    wait4user( GO_BACK_MSG);
	    goto  repeat;
	  case 'c':
	    if ( change_resolution())
	      continue;
	    goto  repeat;
	}
      break;
    }
  if ( keep.sbuf != NULL)
    free( keep.sbuf);
  return( ch);
}


static void
fp_draw_julia( itc, upx, y, real_max, imag_val, real_step, imag_step,
	       real_con, imag_con)
  int	  itc, upx, y;
  double  real_max, imag_val, real_step, imag_step, real_con, imag_con;
{
  int	  iter, x;
  double  real_val, real, imag, real_and_imag, square_sum;

  /* for a description of what's going on here, have a look at the
   * assembly file `julia.S'
   */
  real_max += real_step;
  imag_val -= imag_step;
  do
    {
      imag_val += imag_step;
      real_val= real_max;
      x= upx;
      do
	{
	  real_val -= real_step;
	  real= real_val;
	  imag= imag_val;
	  for ( iter= itc;  ;  )
	    {
	      real_and_imag= real+ imag;
	      if ( ( real *= real) >= 4.0 || ( imag *= imag) >= 4.0
		   || ( square_sum= real+ imag) >= 4.0)
		{
		  iter= ( ( itc- iter) & 127)+ 1;
		  break;
		}
	      if ( --iter == 0)
		break;
	      real += real_con- imag;
	      imag= real_and_imag* real_and_imag- square_sum+ imag_con;
	    }
	  vga_setcolor( iter);
	  vga_drawpixel( x, y);
	}
      while ( --x >= 0);
    }
  while ( --y >= 0);
}

  /* --- start of various subroutines --- */

#define	ERROR_MSG	"\aError: %s\npress any key to choose again..."

static int
do_zoom( funcp, stage, iter, real_min, real_max, imag_min, imag_max)
  int	  ( *funcp) P( (int, int, double, double, double, double) ),
	  stage, iter;
  double  real_min, real_max, imag_min, imag_max;
{
  int	  xs, xe, ys, ye, hold;
  double  scale;

  hold= modep->x/ 4;
  xs= get_X_value( hold);
  for ( xe= modep->x- hold;  ( xe= get_X_value( xe)) == xs;  )
    {
      printf( ERROR_MSG, "right X boundary equals left boundary");
      wait4user( NULL);
    }
  if ( xs > xe)
    {
      hold= xs;  xs= xe;  xe= hold;
    }
  hold= modep->y/ 4;
  ys= get_Y_value( hold);
  for ( ye= modep->y- hold;  ( ye= get_Y_value( ye)) == ys;  )
    {
      printf( ERROR_MSG, "lower Y boundary equals upper boundary");
      wait4user( NULL);
    }
  if ( ys > ye)
    {
      hold= ys;  ys= ye;  ye= hold;
    }
  iter= get_iter_count( iter);
  scale= fabs( real_max- real_min)/ modep->x;
  real_min += scale* xs;
  real_max -= scale* ( modep->x- ( xe+ 1));
  scale= fabs( imag_max- imag_min)/ modep->y;
  imag_min += scale* ( modep->y- ( ye+ 1));
  imag_max -= scale* ys;
  return( ( *funcp)( stage+ 1, iter, real_min, real_max, imag_min, imag_max));
}


#define	ANY_KEY_MSG	"continue"


static unsigned char
change_resolution()
{
  int  i, ch;

  printf( "\nSelect Resolution\n=================\n\n");
  for ( i= 0;  i < MAX_MODES;  i++)
    printf( " %c%c:%5dx%d %s\n",
	    &mode_tab[ i] == modep ? '*' : ' ', i+ '1',
	    mode_tab[ i].x, mode_tab[ i].y,
	    &"unavailable"[ mode_tab[ i].mode < 0 ? 0 : 2]);
  printf( "\nchoose key > ");
  fflush( stdout);
  vga_flip();
  while ( ( ch= fgetc( stdin)) < '1' || ch > MAX_MODES+ '0')
    {
      fputc( '\a', stdout);
      fflush( stdout);
    }
  printf( "%c\n", ch);
  i= ch- '1';
  if ( mode_tab[ i].mode < 0)
    {
      printf( "\aError: resolution %dx%d unavailable - using current (%dx%d)\n",
	      mode_tab[ i].x, mode_tab[ i].y, modep->x, modep->y);
      wait4user( ANY_KEY_MSG);
      return( FALSE);
    }
  vga_flip();
  if ( &mode_tab[ i] == modep)
    return( FALSE);
  init_mode( &mode_tab[ i]);
  return( TRUE);
}


static void
keep_image( kp)
  struct keep  *kp;
{
  int		 x, y;
  unsigned char	 *sp;

  kp->mp= modep;
  if ( ( sp= malloc( modep->x* modep->y)) == NULL)
    {
      printf( "\aWarning: unable to allocate %.2f kB of memory "
	      "- image not marked\n",
	      ( modep->x* modep->y)/ 1024.0);
      wait4user( ANY_KEY_MSG);
    }
  else
    {
      kp->sbuf= sp;
      for ( y= 0;  y < modep->y;  y++)
	for ( x= 0;  x < modep->x;  x++)
	  *sp++= vga_getpixel( x, y);
      printf( "\aimage marked\n");
    }
}


static void
last_image( kp)
  struct keep  *kp;
{
  int		 x, y;
  unsigned char	 *sp;

  if ( modep != kp->mp)
    init_mode( kp->mp);
  vga_setcolor( 0);
  vga_clear();
  sp= kp->sbuf;
  for ( y= 0;  y < modep->y;  y++)
    for ( x= 0;  x < modep->x;  x++)
      {
	vga_setcolor( *sp++);
	vga_drawpixel( x, y);
      }
}


static int
get_iter_count( iter)
  int  iter;
{
  int	new_iter, ch;
  char	msg_buf[ 80];

  for ( ; ; )
    {
      new_iter= 0;
      while ( ( ch= vga_getch()) <= '9' && ch >= '0')
	new_iter= 10* new_iter+ ( ch- '0');
      if ( new_iter == 0)
	{
	  new_iter= iter;
	  break;
	}
      if ( new_iter >= iter)
	break;
      sprintf( msg_buf, "iteration count %d is below previous one (%d)",
	       new_iter, iter);
      printf( ERROR_MSG, msg_buf);
      wait4user( NULL);
    }
  return( new_iter);
}


#define	AXIS_MSG	\
	"\nSelect %c Axis Help\n==================\n\n" \
	"%s\n" \
	"  `1' .. `0':\t\t\tadjust movement step from 1 to 10\n" \
	"  `<space>' or `<return>':\taccept current axis setting\n" \
	"  `h':\t\t\t\tthis help screen\n\n"

#define	DEFAULT_STEP	5


static int
  step= DEFAULT_STEP;


static int
get_X_value( x)
  int  x;
{
  int		 y, ch, nx;
  unsigned char	 linebuf[ MAX_Y_RESO ];

  for ( step= DEFAULT_STEP;  ;  x= nx)
    {
      for ( y= 0;  y < modep->y;  y++)
	{
	  linebuf[ y]= vga_getpixel( x, y);
	  vga_setcolor( 129);
	  vga_drawpixel( x, y);
	}
    repeat:
      switch ( ch= vga_getch())
	{
	  case 'h':
	    printf( AXIS_MSG, 'X',
		    "  `l' or `<left>':\t\tmove axis left\n"
		    "  `r' or `<right>':\t\tmove axis right");
	    wait4user( GO_BACK_MSG);
	    goto  repeat;
	  case '1':  case '2':  case '3':  case '4':  case '5':
	  case '6':  case '7':  case '8':  case '9':
	    step= ch- '0';
	    goto  repeat;
	  case '0':
	    step= 10;
	    goto  repeat;
	  case ' ':  case '\n':  case '\r':
	    return( x);
	  case 27:
	    if ( vga_getch() == '[')
	      if ( ( ch= vga_getch()) == 'D')
		{
	  case 'l':
		  if ( ( nx= x- step) < 0)
		    nx= 0;
		  break;
		}
	      else if ( ch == 'C')
		{
	  case 'r':
		  if ( ( nx= x+ step) >= modep->x)
		    nx= modep->x- 1;
		  break;
		}
	  default:
	    fputc( '\a', stdout);
	    fflush( stdout);
	    goto  repeat;
	}
      for ( y= 0;  y < modep->y;  y++)
	{
	  vga_setcolor( linebuf[ y]);
	  vga_drawpixel( x, y);
	}
    }
}


static int
get_Y_value( y)
  int  y;
{
  int		 x, ch, ny;
  unsigned char	 linebuf[ MAX_X_RESO ];

  for ( step= DEFAULT_STEP;  ;  y= ny)
    {
      for ( x= 0;  x < modep->x;  x++)
	{
	  linebuf[ x]= vga_getpixel( x, y);
	  vga_setcolor( 129);
	  vga_drawpixel( x, y);
	}
    repeat:
      switch ( ch= vga_getch())
	{
	  case 'h':
	    printf( AXIS_MSG, 'Y',
		    "  `u' or `<up>':\t\tmove axis up\n"
		    "  `d' or `<down>':\t\tmove axis down");
	    wait4user( GO_BACK_MSG);
	    goto  repeat;
	  case '1':  case '2':  case '3':  case '4':  case '5':
	  case '6':  case '7':  case '8':  case '9':
	    step= ch- '0';
	    goto  repeat;
	  case '0':
	    step= 10;
	    goto  repeat;
	  case ' ':  case '\n':  case '\r':
	    return( y);
	  case 27:
	    if ( vga_getch() == '[')
	      if ( ( ch= vga_getch()) == 'A')
		{
	  case 'u':
		  if ( ( ny= y- step) < 0)
		    ny= 0;
		  break;
		}
	      else if ( ch == 'B')
		{
	  case 'd':
		  if ( ( ny= y+ step) >= modep->y)
		    ny= modep->y- 1;
		  break;
		}
	  default:
	    fputc( '\a', stdout);
	    fflush( stdout);
	    goto  repeat;
	}
      for ( x= 0;  x < modep->x;  x++)
	{
	  vga_setcolor( linebuf[ x]);
	  vga_drawpixel( x, y);
	}
    }
}


static void
wait4user( prompt)
  const char  *prompt;
{
  if ( prompt != NULL)
    printf( "press any key to %s...", prompt);
  fflush( stdout);
  vga_flip();
  fgetc( stdin);
  vga_flip();
  fputc( '\n', stdout);
}


static void
print_time( stage, st)
  int	   stage;
  clock_t  st;
{
  double  tm;
  int	  days, hrs, min;

  printf( "\ttime used for stage %d: ", stage);
  tm= st/ (double) CLOCKS_PER_SEC;
  if ( tm < 60.0)
    printf( "%.2f seconds\n", tm);
  else if ( tm < 60.0* 60.0)
    {
      min= tm/ 60.0;
      tm -= min* 60.0;
      printf( "%d minute%s %2.2f seconds\n",
	      min, &"s"[ min == 1 ? 1 : 0], tm);
    }
  else if ( tm < 24.0* 60.0* 60.0)
    {
      hrs= tm/ ( 60.0* 60.0);
      tm -= hrs* ( 60.0* 60.0);
      min= tm/ 60.0;
      tm -= min* 60.0;
      printf( "%d hour%s %2d minute%s %2.2f seconds\n",
	      hrs, &"s"[ hrs == 1 ? 1 : 0],
	      min, &"s"[ min == 1 ? 1 : 0], tm);
    }
  else
    {
      days= tm/ ( 24.0* 60.0* 60.0);
      tm -= days* ( 24.0* 60.0* 60.0);
      hrs= tm/ ( 60.0* 60.0);
      tm -= hrs* ( 60.0* 60.0);
      min= tm/ 60.0;
      tm -= min* 60.0;
      printf( "%d day%s %2d hour%s %2d minute%s %2.2f seconds\n",
	      days, &"s"[ days == 1 ? 1 : 0],
	      hrs, &"s"[ hrs == 1 ? 1 : 0],
	      min, &"s"[ min == 1 ? 1 : 0], tm);
    }
}


struct col_entry {
    unsigned char  red,
		   green,
		   blue;
  };

/* below is the color palette table - took me a while to find the
 * appropriate values... :-|
 * we'll use 130 colors, black, which is used for points inside the
 * Mandelbrot set, 128 "speed" colors denoting the speed the value needed
 * to escape and white for drawing the zoom-box-lines
 */

static struct col_entry
  pal_tab[ 1+ 128+ 1]= {
      {  0,  0,  0 },	/* black */
      {  0, 16, 16 },	/* dark cyan -> light cyan */
      {  0, 17, 17 },
      {  0, 18, 18 },
      {  0, 19, 19 },
      {  0, 20, 20 },
      {  0, 21, 21 },
      {  0, 22, 22 },
      {  0, 23, 23 },
      {  0, 24, 24 },
      {  0, 25, 25 },
      {  0, 26, 26 },
      {  0, 27, 27 },
      {  0, 28, 28 },
      {  0, 29, 29 },
      {  0, 30, 30 },
      {  0, 31, 31 },
      {  0, 32, 32 },
      {  0, 33, 33 },
      {  0, 34, 34 },
      {  0, 35, 35 },
      {  0, 36, 36 },
      {  0, 37, 37 },
      {  0, 38, 38 },
      {  0, 39, 39 },
      {  0, 40, 40 },
      {  0, 41, 41 },
      {  0, 42, 42 },
      {  0, 43, 43 },
      {  0, 44, 44 },
      {  0, 45, 45 },
      {  0, 46, 46 },
      {  0, 47, 47 },
      {  0, 48, 48 },
      {  1, 48, 48 },	/* light cyan -> bright violet */
      {  2, 48, 48 },
      {  3, 48, 48 },
      {  4, 48, 48 },
      {  5, 48, 48 },
      {  6, 48, 48 },
      {  7, 48, 48 },
      {  8, 48, 48 },
      {  9, 48, 48 },
      { 10, 48, 48 },
      { 11, 48, 48 },
      { 12, 48, 48 },
      { 13, 48, 48 },
      { 14, 48, 48 },
      { 15, 48, 48 },
      { 16, 48, 48 },
      { 17, 48, 48 },
      { 18, 48, 48 },
      { 19, 48, 48 },
      { 20, 48, 48 },
      { 21, 48, 48 },
      { 22, 48, 48 },
      { 23, 48, 48 },
      { 24, 48, 48 },
      { 25, 48, 48 },
      { 26, 48, 48 },
      { 27, 48, 48 },
      { 28, 48, 48 },
      { 29, 48, 48 },
      { 30, 48, 48 },
      { 31, 48, 48 },
      { 32, 48, 48 },
      { 32, 47, 48 },	/* bright violet -> dark violet */
      { 32, 46, 48 },
      { 32, 45, 48 },
      { 32, 44, 48 },
      { 32, 43, 48 },
      { 32, 42, 48 },
      { 32, 41, 48 },
      { 32, 40, 48 },
      { 32, 39, 48 },
      { 32, 38, 48 },
      { 32, 37, 48 },
      { 32, 36, 48 },
      { 32, 35, 48 },
      { 32, 34, 48 },
      { 32, 33, 48 },
      { 32, 32, 48 },
      { 32, 31, 48 },
      { 32, 30, 48 },
      { 32, 29, 48 },
      { 32, 28, 48 },
      { 32, 27, 48 },
      { 32, 26, 48 },
      { 32, 25, 48 },
      { 32, 24, 48 },
      { 32, 23, 48 },
      { 32, 22, 48 },
      { 32, 21, 48 },
      { 32, 20, 48 },
      { 32, 19, 48 },
      { 32, 18, 48 },
      { 32, 17, 48 },
      { 32, 16, 48 },
      { 31, 16, 47 },	/* dark violet -> dark cyan */
      { 30, 16, 46 },
      { 29, 16, 45 },
      { 28, 16, 44 },
      { 27, 16, 43 },
      { 26, 16, 42 },
      { 25, 16, 41 },
      { 24, 16, 40 },
      { 23, 16, 39 },
      { 22, 16, 38 },
      { 21, 16, 37 },
      { 20, 16, 36 },
      { 19, 16, 35 },
      { 18, 16, 34 },
      { 17, 16, 33 },
      { 16, 16, 32 },
      { 15, 16, 31 },
      { 14, 16, 30 },
      { 13, 16, 29 },
      { 12, 16, 28 },
      { 11, 16, 27 },
      { 10, 16, 26 },
      {  9, 16, 25 },
      {  8, 16, 24 },
      {  7, 16, 23 },
      {  6, 16, 22 },
      {  5, 16, 21 },
      {  4, 16, 20 },
      {  3, 16, 19 },
      {  2, 16, 18 },
      {  1, 16, 17 },
      { 63, 63, 63 }	/* white */
    };

#define	MAX_PAL_COLS	( sizeof( pal_tab)/ sizeof( pal_tab[ 0]))


static void
init_mode( mp)
  struct mode_info  *mp;
{
  int  i;

  modep= mp;
  vga_setmode( mp->mode);
  for ( i= 0;  i < MAX_PAL_COLS;  i++)
    vga_setpalette( i, pal_tab[ i].red, pal_tab[ i].green, pal_tab[ i].blue);
}


static void
save_image( type, iter, real_min, real_max, imag_min, imag_max)
  char	  *type;
  int	  iter;
  double  real_min, real_max, imag_min, imag_max;
{
  int		    x, y, ch;
  unsigned char	    linebuf[ MAX_X_RESO ], savebuf[ 3* MAX_X_RESO ], *sp;
  struct col_entry  *colp;
  FILE		    *fp;
  char		    *p, fname[ FILENAME_MAX ];
  time_t	    act_time;

  vga_flip();
  printf( "\nSave Current Image\n==================\n\n");
  if ( ( p= getcwd( NULL, FILENAME_MAX)) == NULL)
    printf( "\aError: unable to get %s\n", "current working directory");
  else
    {
      printf( "%s: `%s'\n", "current working directory", p);
      free( p);
    }
  printf( "enter file name > ");
  fflush( stdout);
  p= fname;
  while ( ( ch= fgetc( stdin)) != EOF && ch != '\n')
    {
      if ( isalnum( ch) || ispunct( ch))
	*p++= ch;
      else if ( ch == 127 && p > fname)
	{
	  --p;
	  printf( "\b ");
	  ch= '\b';
	}
      else
	{
	  if ( ch == 27)
	    {
	      fgetc( stdin);
	      fgetc( stdin);
	    }
	  ch= '\a';
	}
      fputc( ch, stdout);
      fflush( stdout);
    }
  if ( strcpy( p, ".ppm") == fname)
    printf( "\nimage not saved - ");
  else if ( ( fp= fopen( fname, "w")) == NULL)
    printf( "\n\aError: unable to open file `%s' for writing%s",
	    fname, "\nimage not saved - ");
  else
    {
      printf( "\nsaving image to file `%s' ...", fname);
      fflush( stdout);
      time( &act_time);
      fprintf( fp, "P6\n# Generated by:\t%s %s\n#           at:\t%s",
	       progname, VERSION, ctime( &act_time));
      fprintf( fp, "# %s\t%d iterations\n", type, iter);
      if ( *type == 'J')
	fprintf( fp, "#\t\tconstant( %1.9f, %1.9f)\n", jreal_con, jimag_con);
      fprintf( fp, "#\t\t%s( %1.9f, %1.9f)\n", "real", real_min, real_max);
      fprintf( fp, "#\t\t%s( %1.9f, %1.9f)\n", "imag", imag_min, imag_max);
      fprintf( fp, "%d %d 144\n", modep->x, modep->y);
      vga_flip();
      for ( y= 0;  y < modep->y;  y++)
	{
	  sp= savebuf;
	  for ( x= 0;  x < modep->x;  x++)
	    {
	      colp= &pal_tab[ linebuf[ x]= vga_getpixel( x, y)];
	      vga_setcolor( 129);
	      vga_drawpixel( x, y);
	      *sp++= 3* colp->red;
	      *sp++= 3* colp->green;
	      *sp++= 3* colp->blue;
	    }
	  fwrite( savebuf, sizeof( unsigned char), sp- savebuf, fp);
	  for ( x= 0;  x < modep->x;  x++)
	    {
	      vga_setcolor( linebuf[ x]);
	      vga_drawpixel( x, y);
	    }
	}
      vga_flip();
      printf( " done\n");
      fclose( fp);
    }
  printf( "press any key to %s...", ANY_KEY_MSG);
  fflush( stdout);
  fgetc( stdin);
  fputc( '\n', stdout);
  vga_flip();
}
