/* This routine is slightly modified from `mandel.S' in order to compute
 * the Julia set...
 *
 * Algorithm:
 *   z = z^2 + c { z and c are complex numbers }
 *   z starts with the coordinates x + yi
 *   c are the constant Mandelbrot coordinates cx + cyi
 *
 * void
 * asm_draw_julia( itc, upx, y, r_max, i_val, d_real, d_imag, r_con, i_con)
 *   int  itc,		* iteration count
 *	  upx,		* column (right-most)
 *	  y,		* row (bottom-line)
 *	  r_max,	* real value
 *	  i_val,	* imag value
 *	  d_real,	* delta real
 *	  d_imag,	* delta imag
 *	  r_con,	* real const
 *	  i_con;	* imag const
 *
 * stack layout:
 *   32+36 == 68(%esp) -> i_con
 *   32+32 == 64(%esp) -> r_con
 *   32+28 == 60(%esp) -> d_imag
 *   32+24 == 56(%esp) -> d_real
 *   32+20 == 52(%esp) -> i_val
 *   32+16 == 48(%esp) -> r_max
 *   32+12 == 44(%esp) -> y
 *   32+08 == 40(%esp) -> upx
 *   32+04 == 36(%esp) -> itc			* >  parameters
 *   32+00 == 32(%esp) -> return address	* == link
 *   32-04 == 28(%esp) -> %ebx			* <  caller-save registers
 *   32-08 == 24(%esp) -> %ebp
 *   32-12 == 20(%esp) -> %edi
 *   32-16 == 16(%esp) -> %esi
 *   32-20 == 12(%esp) -> r_val			* <  local variables
 *   32-24 ==  8(%esp) -> x
 *   32-28 ==  4(%esp) -> par.y			* <  function parameters
 *   32-32 ==   (%esp) -> par.x or color
 *
 * register usage:
 *   %ebx -> real
 *   %ebp -> imag
 *   %ecx -> iter
 *   %edi -> real con
 *   %esi -> imag con
 *   %eax, %edx -> needed for multiplication #-(
 *
 * Note:  while reading the comments below keep in mind, that instructions
 *	  are scheduled, where possible, to omit penalty clock's on a 486er
 */

.text
#ifndef	__ELF__
	.align	4
.globl	_asm_draw_julia
_asm_draw_julia:
#else
	.align	16
.globl	asm_draw_julia
	.type	asm_draw_julia,@function
asm_draw_julia:
#endif
	pushl	%ebx		/* preserve caller-save registers */
	pushl	%ebp
	pushl	%edi
	pushl	%esi
	subl	$16,%esp	/* 2 int's and 2 parameters needed */
	movl	64(%esp),%edi	/* %edi= r_con */
	movl	68(%esp),%esi	/* %esi= i_con */
	movl	60(%esp),%eax	/* i_val -= d_imag; */
	movl	56(%esp),%ecx	/* r_max += d_real; */
	subl	%eax,52(%esp)
	addl	%ecx,48(%esp)
y_loop:
	movl	60(%esp),%eax	/* i_val += d_imag; */
	movl	40(%esp),%ecx	/* x= upx */
	movl	48(%esp),%edx	/* r_val= r_max; */
	addl	%eax,52(%esp)
	movl	%ecx,8(%esp)
	movl	%edx,12(%esp)
x_loop:
	movl	12(%esp),%ebx	/* %ebx= real= rval -= d_real */
	movl	52(%esp),%ebp	/* %ebp= imag= i_val */
	subl	56(%esp),%ebx
	movl	36(%esp),%ecx	/* %ecx= iter= itc */
	movl	%ebx,12(%esp)
pre_iter_begin:
	movl	%ebx,%eax	/* pre-evaluate the sequence located
				 * between the two labels
				 * `iter_begin' and `iter_exit'
				 */
	imull	%eax
	sall	$3,%edx
	jo	iter_exit
	jc	iter_exit
	movl	%ebp,%eax
	addl	%ebx,%ebp
	movl	%edx,%ebx
	imull	%eax
	sall	$3,%edx
	jo	iter_exit
	jc	iter_exit
	addl	%ebx,%edx
	jo	iter_exit
iter_repeat:
	decl	%ecx		/* if ( --iter == 0)  break; */
	jz	color_zero	/* assume it's inside, so "paint it black"
				 * ahem, any similarity with that famous
				 * Stones' song is purely random ;-)
				 */
	movl	%ebp,%eax	/* %eax= real+ imag
				 *
				 * %ebx= real^2- imag^2+ imag_con
				 *     = 2* real^2- ( real^2+ imag^2)
				 *	 + imag_con
				 */
	leal	(%edi,%ebx,2),%ebx
	movl	%esi,%ebp
	subl	%edx,%ebx
				/* %ebp= 2* real* imag+ imag_con
				 *     = imag_con- ( real^2+ imag^2)
				 *	 + ( real+ imag)^2
				 */
	subl	%edx,%ebp
	imull	%eax		/* ( real+ imag)^2 */
	leal	(%ebp,%edx,8),%ebp
iter_begin:
	movl	%ebx,%eax	/* %eax= real */
	imull	%eax
	sall	$3,%edx		/* if ( real^2 >= 4)  break;
				 * correctly this should be a
				 * `shldl $3,%eax,%edx', but
				 * linux-gas v2.5.2.6 doesn't recognize
				 * the immediate parameter as a valid one #-(
				 * the only solution would be to setup %ecx
				 * with the value 3, but that would mean this
				 * register isn't free as the iteration-counter
				 * any longer - not worth for the extra
				 * precision of 5.587935448 * 10^-9 IMHO
				 */
	jo	iter_exit
	jc	iter_exit	/* this 2nd j<cc> is necessary, because
				 * Intel chip designer's doesn't know the
				 * difference between an arithmetical and
				 * a logical left-shift %-6
				 * (hint: it's a matter of how to set the
				 *	  V condition bit)
				 */
	movl	%ebp,%eax	/* %eax= imag */
	addl	%ebx,%ebp	/* %ebp= real+ imag */
	movl	%edx,%ebx	/* %ebx= real^2 */
	imull	%eax
	sall	$3,%edx		/* if ( imag^2 >= 4)  break; */
	jo	iter_exit
	jc	iter_exit	/* see comments above ! */
	addl	%ebx,%edx	/* if ( real^2+ imag^2 >= 4)  break; */
	jno	iter_repeat
iter_exit:
	negl	%ecx		/* iter= itc- iter */
	addl	36(%esp),%ecx	/* we'll use 129 colors, black for all values
				 * inside the Mandelbrot set and 128 "speed"
				 * measuring colors
				 */
	andl	$127,%ecx	/* color= ( iter % 128)+ 1; */
	incl	%ecx		/* skip over black */
color_zero:
	movl	44(%esp),%eax	/* %eax= y */
	movl	8(%esp),%ebp	/* %ebp= x */
	movl	%eax,4(%esp)	/* "push" parameter y */
	movl	%ecx,(%esp)	/* vga_setcolor( color); */
#ifndef	__ELF__
	call	_vga_setcolor
#else
	call	vga_setcolor
#endif
	movl	%ebp,(%esp)	/* vga_drawpixel( x, y); */
#ifndef	__ELF__
	call	_vga_drawpixel
#else
	call	vga_drawpixel
#endif
x_end:
	decl	8(%esp)		/* do..while ( --x >= 0); */
	jge	x_loop
y_end:
	decl	44(%esp)	/* do..while ( --y >= 0); */
	jge	y_loop
	addl	$16,%esp	/* remove 2 parameters and 2 int's */
	popl	%esi		/* retrieve caller-save registers */
	popl	%edi
	popl	%ebp
	popl	%ebx
	ret
#ifdef	__ELF__
end:
	.size	asm_draw_julia,end-asm_draw_julia
#endif
